/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                               interpol.cpp
 *
 *  Interpolation/concentration averaging object "InterpolObj" 
 *  (appear in script as "Ca[x,y,z]" or "Ca[]")
 *
 *  Array of interpolation objects class "InterpolArray"
 *  InterpolArray is a member of the "KineticObj" defined in gate.h/gate.cpp
 *
 ******************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "grid.h"
#include "field.h"
#include "interpol.h"

//**************************************************************************

double InterpolObj::Evaluate(double t)
   {
     if (*tptr != newtime)   // field has been updated since last interpolation evaluation
       {
       oldtime = newtime;
       oldres = newres;
       newtime = *tptr;

       newres  = average ? Field->average() :
                *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
                *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;
       }

     if (newtime > oldtime) // temporal extrapolation using values at two different times  
       result = ( (newtime - t) * oldres + (t - oldtime) * newres ) / (newtime - oldtime);
     else result = newres;
     
     return result;
   }
//**************************************************************************

void InterpolObj::reset()
   {
   if (average) return;

   result = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
            *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;

   oldres = newres = result;
   oldtime = newtime = *tptr;   
   }

//*******************************************************************************
//*******************************************************************************

InterpolObj::InterpolObj(TokenString &Param, long p0, FieldObj *FO)
   {
     double *xcoord = FieldObj::Grid->xcoord; int xsize = FieldObj::Grid->xsize;
     double *ycoord = FieldObj::Grid->ycoord; int ysize = FieldObj::Grid->ysize;
     double *zcoord = FieldObj::Grid->zcoord; int zsize = FieldObj::Grid->zsize;
     int xysize = FieldObj::Grid->xysize;


   double x=0.0, y=0.0, z=0.0;
   long p = p0, pLast;
   average = false;   
   char xyz[128];
   
   Field = FO;
   double *field = (*FO)();
   tptr = &(FO->Time); 
 
   if ( Param.equal(p + 1, "]") )  {
     strcpy(xyz, Param[p-1]); strcat( xyz, "[]" );
     average = true;
     Evaluate(0.0);
     pLast = p;
     }
   else {

   if ( DIMENSIONALITY > 2 ) {
     x = ExpressionObj(Param, p + 1, "", 0, &pLast).Evaluate();
     if ( !Param.equal(pLast + 1, ",") ) 
       Param.errorMessage(pLast + 1, 0, "Bad coordinate specifier or a missing comma");
     p = pLast + 1;
   }

   if ( DIMENSIONALITY > 1 ) {
     y = ExpressionObj(Param, p + 1, "", 0, &pLast).Evaluate();
     if ( !Param.equal(pLast + 1, ",") ) 
       Param.errorMessage(pLast + 1, 0, "Bad coordinate specifier or a missing comma");
     p = pLast + 1;
   }

   z = ExpressionObj(Param, p + 1, "", 0, &pLast).Evaluate();
   if ( !Param.equal(pLast + 1, "]") ) 
       Param.errorMessage(pLast + 1, 0, "Bad coordinate specifier or a missing closing \"]\"");
   
 
   if ( DIMENSIONALITY == 1 ) 
      { x = z; y = z = 0.0; sprintf(xyz,"%s[%g]", FO->ID, x); }
   else if ( DIMENSIONALITY == 2 ) 
           { x = y; y = z; z = 0.0; sprintf(xyz,"%s[%g,%g]", FO->ID, x, y); }
   else sprintf(xyz,"%s[%g,%g,%g]", FO->ID, x, y, z);

   if ( !FieldObj::Grid->in_bounds(x, y, z) ) {
     char *s;
     if ( DIMENSIONALITY > 2 ) s = makeMessage("Location [%g,%g,%g] is out of bounds", x, y, z);
     else if ( DIMENSIONALITY > 1 ) s = makeMessage("Location [%g,%g] is out of bounds", x, y);
     else s = makeMessage("Location [%g] is out of bounds", x);
     Param.errorMessage(p0, s);
   }
   }   
   
   ID = StrCpy( xyz );

   token_ptr_source = &(Param.token_ptr[p0-1]);   // restore at destruction time
   token_ptr_target = Param.token_ptr[p0-1];      //  
   Param.token_ptr[p0-1] = ID;                    // *** hijack the token pointer
   Param.deleteToken(p0, pLast + 1);

   oldtime = newtime = *tptr;   

   if (average)  {  result = Field->average();
                    oldres = newres = result;
                    return;
   }

   double p1 = 1.0, p2 = 1.0, p3 = 1.0;
   int    ix = 0, iy = 0, iz = 0;  
   
   while (x >= xcoord[ix] && ix < (xsize - 1) ) ix++;
   if ( DIMENSIONALITY > 1 )  while (y >= ycoord[iy] && iy < (ysize - 1) ) iy++;
   if ( DIMENSIONALITY > 2 )  while (z >= zcoord[iz] && iz < (zsize - 1) ) iz++;

   int opt = 0;

   p1 = (x - xcoord[ix-1]) / (xcoord[ix] - xcoord[ix-1]);
   if ( DIMENSIONALITY > 1) p2 = (y - ycoord[iy-1]) / (ycoord[iy] - ycoord[iy-1]);
   if ( DIMENSIONALITY > 2) p3 = (z - zcoord[iz-1]) / (zcoord[iz] - zcoord[iz-1]);

   f000 = f001 = f010 = f011 = f100 = f101 = f110 = f111 = field + ix + iy * xsize + iz * xysize;

   if ( ix > 0 ) { f000 -= 1;      f001 -= 1;      f010 -= 1 ;     f011 -= 1; }
   if ( iy > 0 ) { f000 -= xsize;  f001 -= xsize;  f100 -= xsize ; f101 -= xsize; }
   if ( iz > 0 ) { f000 -= xysize; f010 -= xysize; f100 -= xysize; f110 -= xysize ; }

   p111 =         p1  *      p2  *      p3 ;
   p000 =    (1 - p1) * (1 - p2) * (1 - p3);
   p100 =         p1  * (1 - p2) * (1 - p3);
   p010 =    (1 - p1) *      p2  * (1 - p3);
   p001 =    (1 - p1) * (1 - p2) *      p3 ;
   p110 =         p1  *      p2  * (1 - p3);
   p101 =         p1  * (1 - p2) *      p3 ;
   p011 =    (1 - p1) *      p2  *      p3 ;
 
   result = *f000 * p000 + *f001 * p001 + *f010 * p010 + *f011 * p011 +   
            *f100 * p100 + *f101 * p101 + *f110 * p110 + *f111 * p111 ;
   
   oldres = newres = result;
   }
   
//*******************************************************************************

InterpolArray::InterpolArray(TokenString &Param, FieldObj *Ca, BufferArray *Bufs)
{
  interpol_num = 0;
  long p0;
  int i, j;

  if (Ca) interpol_num += Param.token2_count(Ca->ID, "[");

  if (Bufs)
    for (j = 0; j < Bufs->buf_num; j++)
        interpol_num += Param.token2_count(Bufs->array[j]->ID, "[");

    array = new InterpolObj *[interpol_num];
    int count = 0;

    if (!interpol_num) return; 

    if (VERBOSE) { fprintf(stdlog,"\n### Interpolating: "); fflush(stdlog); }

    if (Ca) 
      while ( Param.token2_count(Ca->ID, "[") )
        {
        p0 = Param.token2_index(Ca->ID, "[", 1);
        array[count] = new InterpolObj(Param, p0, Ca );
        if (VERBOSE) fprintf(stdlog," %s",array[count]->ID);
        count ++;
        }
      

    if (Bufs)
      for (j = 0; j < Bufs->buf_num; j++)
        while ( Param.token2_count(Bufs->array[j]->ID, "[") )
          {
          p0 = Param.token2_index(Bufs->array[j]->ID, "[", 1);
          array[count] = new InterpolObj(Param, p0, Bufs->array[j] );
          if (VERBOSE) fprintf(stdlog," %s",array[count]->ID);
          count ++;
          }

    if (VERBOSE) fprintf(stdlog,"\n # Compressing interpolation array: ");

    InterpolObj *temp;

    for (i = 0; i < interpol_num - 1; i++ )
      for  (j = i + 1; j < interpol_num; j++ )
	  if ( equal(array[i]->ID, array[j]->ID) )
	    {
	    if (VERBOSE) { fprintf(stdlog,"."); fflush(stdlog); }
            interpol_num--;
	    temp = array[interpol_num]; 
            array[interpol_num] = array[j];
            array[j] = temp;
	    }

    if (VERBOSE) fprintf(stdlog," done\n");

}

//*******************************************************************************

